@@ -14,6 +14,9 @@ from databricks.feature_store.lookup_engine.lookup_dynamodb_engine import (
     LookupDynamoDbEngine,
     AwsAccessKey,
 )
+from databricks.feature_store.lookup_engine.lookup_cosmosdb_engine import (
+    LookupCosmosDbEngine,
+)
 
 __all__ = [
     "LookupEngine",
@@ -21,4 +24,5 @@ __all__ = [
     "LookupMySqlEngine",
     "LookupSqlServerEngine",
     "LookupDynamoDbEngine",
+    "LookupCosmosDbEngine",
 ]
