@@ -4,9 +4,7 @@ import sqlalchemy
 
 from databricks.feature_store.entities.cloud import Cloud
 from databricks.feature_store.entities.online_feature_table import OnlineFeatureTable
-from databricks.feature_store.lookup_engine.lookup_sql_engine import (
-    LookupSqlEngine,
-)
+from databricks.feature_store.lookup_engine.lookup_sql_engine import LookupSqlEngine
 
 
 class LookupMySqlEngine(LookupSqlEngine):
