@@ -3,9 +3,7 @@ import collections
 import sqlalchemy
 
 from databricks.feature_store.entities.online_feature_table import OnlineFeatureTable
-from databricks.feature_store.lookup_engine.lookup_sql_engine import (
-    LookupSqlEngine,
-)
+from databricks.feature_store.lookup_engine.lookup_sql_engine import LookupSqlEngine
 
 
 class LookupSqlServerEngine(LookupSqlEngine):
