@@ -1,36 +1,41 @@
-from collections import defaultdict, Counter
-from typing import List, Dict, Tuple
 import itertools
-import mlflow
 import os
+from collections import Counter, defaultdict
+from typing import Dict, List, Tuple, Union
+
+import mlflow
 import pandas as pd
-from mlflow.utils import databricks_utils
 from mlflow.models.container import SERVING_ENVIRONMENT
 from mlflow.sagemaker import SAGEMAKER_SERVING_ENVIRONMENT
-from typing import List
+from mlflow.utils import databricks_utils
 
+from databricks.feature_store import mlflow_model_constants
 from databricks.feature_store.entities.feature_column_info import FeatureColumnInfo
 from databricks.feature_store.entities.feature_spec import FeatureSpec
 from databricks.feature_store.entities.feature_tables_for_serving import (
     AbstractFeatureTablesForServing,
-    FeatureTablesForServing,
     FeatureTablesForSageMakerServing,
+    FeatureTablesForServing,
 )
 from databricks.feature_store.entities.online_feature_table import (
     AbstractOnlineFeatureTable,
+    OnlineFeatureTable,
     PrimaryKeyDetails,
 )
-from databricks.feature_store.online_lookup_client import OnlineLookupClient
-from databricks.feature_store import mlflow_model_constants
+from databricks.feature_store.entities.store_type import StoreType
 from databricks.feature_store.feature_lookup_version import VERSION
-
-# The provisioner of this model is expected to set an environment variable with the path to a
-# feature_tables_for_serving.dat file.
+from databricks.feature_store.online_lookup_client import (
+    OnlineLookupClient,
+    is_primary_key_lookup,
+    tables_share_dynamodb_access_keys,
+)
 from databricks.feature_store.utils.uc_utils import (
     get_feature_spec_with_full_table_names,
     get_feature_spec_with_reformat_full_table_names,
 )
 
+# The provisioner of this model is expected to set an environment variable with the path to a
+# feature_tables_for_serving.dat file.
 FEATURE_TABLES_FOR_SERVING_FILEPATH_ENV = "FEATURE_TABLES_FOR_SERVING_FILEPATH"
 
 LookupKeyType = Tuple[str, ...]
@@ -83,8 +88,32 @@ class _FeatureStoreModelWrapper:
             self.feature_spec, feature_tables_for_serving
         )
         self._validate_ft_metadata(self.ft_metadata)
+        self.is_model_eligible_for_batch_lookup = (
+            self._is_model_eligible_for_batch_lookup(self.ft_metadata)
+        )
 
-        self.ft_to_lookup_client = self._create_lookup_clients(self.ft_metadata)
+        if self.is_model_eligible_for_batch_lookup:
+            self.batch_lookup_client = self._create_batch_lookup_client(
+                self.ft_metadata
+            )
+        else:
+            self.ft_to_lookup_client = self._create_lookup_clients(self.ft_metadata)
+
+    # true if all feature tables using DynamoDB under same authorization (ie same region and keys)
+    # this is true by default for customers using Sagemaker
+    def _is_model_eligible_for_batch_lookup(
+        self, ft_metadata: Dict[str, _FeatureTableMetadata]
+    ):
+        env = self._get_serving_environment()
+        online_feature_tables = [meta.online_ft for _, meta in ft_metadata.items()]
+        if env == mlflow_model_constants.SAGEMAKER:
+            return is_primary_key_lookup(online_feature_tables)
+        elif env == mlflow_model_constants.DATABRICKS:
+            return tables_share_dynamodb_access_keys(
+                online_feature_tables
+            ) and is_primary_key_lookup(online_feature_tables)
+
+        raise Exception(f"Internal Error: Unexpected serving_environment {env}.")
 
     @staticmethod
     def _load_feature_tables_for_serving(
@@ -137,6 +166,15 @@ class _FeatureStoreModelWrapper:
             )
         return ft_to_lookup_client
 
+    def _create_batch_lookup_client(self, ft_metadata):
+        online_fts = []
+        for ft, meta in ft_metadata.items():
+            online_fts.append(meta.online_ft)
+
+        return OnlineLookupClient(
+            online_fts, serving_environment=self.serving_environment
+        )
+
     def predict(self, df: pd.DataFrame):
         self._validate_input(df)
         model_input_df = self._augment_with_features(df)
@@ -292,10 +330,17 @@ class _FeatureStoreModelWrapper:
             partially_overridden_feature_output_names,
         ) = self._get_overridden_feature_output_names(df)
 
+        pk_dfs = defaultdict(lambda: defaultdict(list))
+        feature_column_infos_to_lookup_dict = defaultdict(lambda: defaultdict(list))
+        lookup_clients = {}
+
         # Query online store(s) for feature values
         ft_meta: _FeatureTableMetadata
         for ft_name, ft_meta in self.ft_metadata.items():
-            lookup_client = self.ft_to_lookup_client[ft_name]
+            if not self.is_model_eligible_for_batch_lookup:
+                lookup_clients[
+                    ft_meta.online_ft.online_feature_table_name
+                ] = self.ft_to_lookup_client[ft_name]
 
             # Iterate through the lookup_keys for this feature table, each of which is used to
             # lookup a list of features
@@ -315,12 +360,36 @@ class _FeatureStoreModelWrapper:
                     # All features were overridden in the input DataFrame
                     continue
 
-                pk_df = self._get_primary_key_df(
+                pk_dfs[ft_meta.online_ft.online_feature_table_name][
+                    lookup_key
+                ] = self._get_primary_key_df(
                     lookup_key, ft_meta.online_ft.primary_keys, df
                 )
-                feature_values_df = self._lookup_and_rename_features(
-                    lookup_client, pk_df, feature_column_infos_to_lookup
-                )
+                feature_column_infos_to_lookup_dict[
+                    ft_meta.online_ft.online_feature_table_name
+                ][lookup_key] = feature_column_infos_to_lookup
+
+        feature_values_dfs = defaultdict(lambda: defaultdict(list))
+        if self.is_model_eligible_for_batch_lookup and pk_dfs:
+            feature_values_dfs = self._batch_lookup_and_rename_features(
+                self.batch_lookup_client, pk_dfs, feature_column_infos_to_lookup_dict
+            )
+        else:
+            for oft_name, pk_df_by_lookup_key in pk_dfs.items():
+                for lookup_key in pk_df_by_lookup_key.keys():
+                    feature_values_dfs[oft_name][
+                        lookup_key
+                    ] = self._lookup_and_rename_features(
+                        lookup_clients[oft_name],
+                        pk_dfs[oft_name][lookup_key],
+                        feature_column_infos_to_lookup_dict[oft_name][lookup_key],
+                    )
+
+        for oft_name, feature_values_df_by_lookup_key in feature_values_dfs.items():
+            for (
+                lookup_key,
+                feature_values_df,
+            ) in feature_values_df_by_lookup_key.items():
                 if feature_values_df.shape[0] != df.shape[0]:
                     raise Exception(
                         f"Internal Error: Expected {df.shape[0]} rows to be looked up from feature "
@@ -396,7 +465,7 @@ class _FeatureStoreModelWrapper:
         lookup_client: OnlineLookupClient,
         primary_key_df: pd.DataFrame,
         feature_column_infos: List[FeatureColumnInfo],
-    ):
+    ) -> pd.DataFrame:
         """
         Looks up features from a single feature table, then renames them. Feature metadata is
          specified via `feature_column_infos`.
@@ -406,9 +475,47 @@ class _FeatureStoreModelWrapper:
         feature_name_to_output_name = {
             fci.feature_name: fci.output_name for fci in feature_column_infos
         }
-
         return feature_values.rename(feature_name_to_output_name, axis=1)
 
+    def _batch_lookup_and_rename_features(
+        self,
+        batch_lookup_client: OnlineLookupClient,
+        primary_key_dfs: Dict[str, Dict[LookupKeyType, pd.DataFrame]],
+        feature_column_infos_dict: Dict[
+            str, Dict[LookupKeyType, List[FeatureColumnInfo]]
+        ],
+    ) -> Dict[str, Dict[LookupKeyType, pd.DataFrame]]:
+        """
+        Looks up features from all the feature tables in batch, then renames them. Feature metadata
+         is specified via `feature_column_infos`.
+        """
+        feature_names = defaultdict(lambda: defaultdict(list))
+        feature_name_to_output_names = defaultdict(lambda: defaultdict(list))
+        for (
+            oft_name,
+            feature_column_infos_by_lookup_key,
+        ) in feature_column_infos_dict.items():
+            for (
+                lookup_key,
+                feature_column_infos,
+            ) in feature_column_infos_by_lookup_key.items():
+                feature_names[oft_name][lookup_key] = [
+                    fci.feature_name for fci in feature_column_infos
+                ]
+                feature_name_to_output_names[oft_name][lookup_key] = {
+                    fci.feature_name: fci.output_name for fci in feature_column_infos
+                }
+        feature_values = batch_lookup_client.batch_lookup_features(
+            primary_key_dfs, feature_names
+        )
+
+        for oft_name, feature_values_by_lookup_key in feature_values.items():
+            for lookup_key, _ in feature_values_by_lookup_key.items():
+                feature_values[oft_name][lookup_key] = feature_values[oft_name][
+                    lookup_key
+                ].rename(feature_name_to_output_names[oft_name][lookup_key], axis=1)
+        return feature_values
+
     def _resolve_online_stores(
         self,
         feature_tables: List[str],
