@@ -1,24 +1,26 @@
+import os
+from typing import Dict, List, Optional, Tuple, Union
+
+import pandas as pd
+
 from databricks.feature_store.entities.online_feature_table import (
     AbstractOnlineFeatureTable,
     OnlineFeatureTable,
     OnlineFeatureTableForSageMakerServing,
 )
+from databricks.feature_store.entities.query_mode import QueryMode
 from databricks.feature_store.entities.store_type import StoreType
 from databricks.feature_store.lookup_engine import (
     AwsAccessKey,
+    LookupCosmosDbEngine,
+    LookupDynamoDbEngine,
+    LookupEngine,
     LookupMySqlEngine,
     LookupSqlEngine,
     LookupSqlServerEngine,
-    LookupEngine,
-    LookupDynamoDbEngine,
-    LookupCosmosDbEngine,
 )
 from databricks.feature_store.mlflow_model_constants import DATABRICKS, SAGEMAKER
 
-from typing import List, Tuple, Optional
-import pandas as pd
-import os
-
 # The provisioner of this model is expected to set the following environment variable for each
 # feature table if the feature store is SQL based:
 #  (1) <online_store_for_serving.read_secret_prefix>_USER
@@ -33,12 +35,18 @@ SECRET_ACCESS_KEY_SUFFIX = "_SECRET_ACCESS_KEY"
 # For Cosmos DB the following variable should be set:
 #  (1) <online_store_for_serving.read_secret_prefix>_AUTHORIZATION_KEY
 AUTHORIZATION_KEY_SUFFIX = "_AUTHORIZATION_KEY"
+LookupKeyType = Tuple[str, ...]
 
 
 def generate_lookup_sql_engine(
-    online_feature_table: OnlineFeatureTable,
+    online_feature_table: Union[
+        AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
+    ],
     creds: Tuple[str, str],
 ) -> LookupSqlEngine:
+    if isinstance(online_feature_table, List):
+        raise Exception(f"Internal error: Batch lookup is unsupported in SQL Engine.")
+
     ro_user, ro_password = creds
     if online_feature_table.online_store.store_type == StoreType.SQL_SERVER:
         return LookupSqlServerEngine(online_feature_table, ro_user, ro_password)
@@ -46,7 +54,9 @@ def generate_lookup_sql_engine(
 
 
 def generate_lookup_dynamodb_engine(
-    online_feature_table: AbstractOnlineFeatureTable,
+    online_feature_table: Union[
+        AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
+    ],
     creds: Optional[Tuple[str, str]],
 ) -> LookupDynamoDbEngine:
     if creds:
@@ -61,8 +71,13 @@ def generate_lookup_dynamodb_engine(
 
 
 def generate_lookup_cosmosdb_engine(
-    online_feature_table: OnlineFeatureTable, creds: str
+    online_feature_table: Union[
+        AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
+    ],
+    creds: str,
 ) -> LookupCosmosDbEngine:
+    if isinstance(online_feature_table, List):
+        raise Exception(f"Internal error: Batch lookup is unsupported in Cosmos DB.")
     return LookupCosmosDbEngine(online_feature_table, authorization_key=creds)
 
 
@@ -89,9 +104,71 @@ def load_credentials_from_env(online_ft: OnlineFeatureTable):
         )
 
 
+# TODO(ML-26146): refactor this method and above helpers to appropriate seperate util files
+def tables_share_dynamodb_access_keys(
+    online_feature_tables: List[AbstractOnlineFeatureTable],
+):
+    # check region and credentials to see if all feature tables are using same DynamoDB accesss keys
+    region, creds = None, None
+    for ft in online_feature_tables:
+        if ft.online_store.store_type != StoreType.DYNAMODB:
+            return False
+        elif region == None:
+            region = ft.online_store.extra_configs.region
+            creds = load_credentials_from_env(ft)
+        elif (
+            region != ft.online_store.extra_configs.region
+            or load_credentials_from_env(ft) != creds
+        ):
+            return False
+    return True
+
+
+def is_primary_key_lookup(
+    online_feature_tables: List[AbstractOnlineFeatureTable],
+):
+    for oft in online_feature_tables:
+        mode = oft.online_store.query_mode
+        if mode == QueryMode.RANGE_QUERY:
+            return False
+        elif mode == QueryMode.PRIMARY_KEY_LOOKUP:
+            continue
+        else:
+            raise ValueError(f"Unsupported query mode: {mode}")
+    return True
+
+
+def check_instance_type(
+    online_feature_table: Union[
+        AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
+    ],
+    cls: Union[OnlineFeatureTable, OnlineFeatureTableForSageMakerServing],
+    serving_environment: str,
+):
+    if isinstance(online_feature_table, cls):
+        return True
+    elif isinstance(online_feature_table, List):
+        for oft in online_feature_table:
+            if not isinstance(oft, cls):
+                raise Exception(
+                    f"Internal Error: Expected {cls} when serving_environment={serving_environment}"
+                    f". Got {type(oft)}."
+                )
+        return True
+
+    raise Exception(
+        f"Internal Error: Expected {cls} when serving_environment={serving_environment}"
+        f". Got {type(online_feature_table)}."
+    )
+
+
 class OnlineLookupClient:
     def __init__(
-        self, online_feature_table: AbstractOnlineFeatureTable, serving_environment: str
+        self,
+        online_feature_table: Union[
+            AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
+        ],
+        serving_environment: str,
     ):
         self.lookup_engine = self._generate_lookup_engine(
             online_feature_table, serving_environment
@@ -99,26 +176,24 @@ class OnlineLookupClient:
 
     @staticmethod
     def _generate_lookup_engine(
-        online_feature_table: AbstractOnlineFeatureTable, serving_environment: str
+        online_feature_table: Union[
+            AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
+        ],
+        serving_environment: str,
     ):
         if serving_environment == DATABRICKS:
-            if not isinstance(online_feature_table, OnlineFeatureTable):
-                raise Exception(
-                    "Internal Error: Expected OnlineFeatureTable when serving_environment=DATABRICKS"
-                    f". Got {type(online_feature_table)}."
-                )
+            check_instance_type(
+                online_feature_table, OnlineFeatureTable, serving_environment
+            )
             return OnlineLookupClient._generate_lookup_engine_databricks(
                 online_feature_table
             )
-
         elif serving_environment == SAGEMAKER:
-            if not isinstance(
-                online_feature_table, OnlineFeatureTableForSageMakerServing
-            ):
-                raise Exception(
-                    "Internal Error: Expected OnlineFeatureTableForSageMakerServing when "
-                    f"serving_environment=DATABRICKS. Got {type(online_feature_table)}."
-                )
+            check_instance_type(
+                online_feature_table,
+                OnlineFeatureTableForSageMakerServing,
+                serving_environment,
+            )
             return OnlineLookupClient._generate_lookup_engine_sagemaker(
                 online_feature_table
             )
@@ -129,12 +204,25 @@ class OnlineLookupClient:
 
     @classmethod
     def _generate_lookup_engine_databricks(
-        cls, online_feature_table: OnlineFeatureTable
+        cls,
+        online_feature_table: Union[
+            AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
+        ],
     ):
-        creds = load_credentials_from_env(online_feature_table)
-        if online_feature_table.online_store.store_type == StoreType.DYNAMODB:
+        first_online_feature_table = None
+        if isinstance(online_feature_table, List):
+            if len(online_feature_table) == 0:
+                raise Exception(
+                    f"Internal Error: No feature table passed while creating lookup engine."
+                )
+            first_online_feature_table = online_feature_table[0]
+        else:
+            first_online_feature_table = online_feature_table
+
+        creds = load_credentials_from_env(first_online_feature_table)
+        if first_online_feature_table.online_store.store_type == StoreType.DYNAMODB:
             return generate_lookup_dynamodb_engine(online_feature_table, creds)
-        elif online_feature_table.online_store.store_type == StoreType.COSMOSDB:
+        elif first_online_feature_table.online_store.store_type == StoreType.COSMOSDB:
             return generate_lookup_cosmosdb_engine(online_feature_table, creds)
         else:
             return generate_lookup_sql_engine(online_feature_table, creds)
@@ -145,6 +233,21 @@ class OnlineLookupClient:
     ):
         return generate_lookup_dynamodb_engine(online_feature_table, creds=None)
 
+    def batch_lookup_features(
+        self,
+        lookup_df_dict: Dict[str, Dict[LookupKeyType, pd.DataFrame]],
+        feature_names_dict: Dict[str, Dict[LookupKeyType, List[str]]],
+    ) -> Dict[str, Dict[LookupKeyType, pd.DataFrame]]:
+        """
+        :param lookup_df_dict: dictionary from online feature table name to primary key dfs.
+        :param feature_names_dict:  dictionary from online feature table name to feature names.
+        :return a Dictionary from online feature table name to feature values.
+        """
+        return self.lookup_engine.batch_lookup_features(
+            lookup_df_dict,
+            feature_names_dict,
+        )
+
     def lookup_features(
         self,
         lookup_df: pd.DataFrame,
