@@ -1,5 +1,5 @@
 import json
-from typing import List, Any
+from typing import Any, List
 
 from azure.cosmos.exceptions import CosmosResourceNotFoundError
 
